#pragma once

#include <pybind11/numpy.h>    // IWYU pragma: export
#include <pybind11/pybind11.h> // IWYU pragma: export
#include <pybind11/stl.h>      // IWYU pragma: export

namespace py = pybind11;
